import numpy as np
import sasktran as sk
from sasktran_tir.constants import Boltzmann
from sasktran_tir.util import _atm_reader, _atm_file_path


class ClimatologyAtmosphericState(sk.ClimatologyUserDefined):
    """
    Creates a custom climatology, containing temperature and pressure, for use as an atmospheric state.
    The profiles are from http://eodg.atm.ox.ac.uk/RFM/atm/ and are divided into three folders:
    'FASCODE/ICRCCM Model Atmospheres', 'MIPAS Model Atmospheres (1998)', and 'MIPAS Model Atmospheres (2001)'.
    Each folder contains several '.atm' files, where each file contains a set of profiles describing various
    latitudinal regions and times.

    NOTE: These are one-dimensional atmospheric profiles. Latitude, longitude, and mjd will not affect the returned
          values.
    """
    def __init__(self, atm_folder='fascode', atm_file='std.atm', interp: str='linear', spline: bool=False):
        """
        Create an atmospheric state from data files. By default uses the US Standard Atmosphere.

        Parameters
        ----------
        atm_folder : str, optional
            Name of dataset to use. Valid choices are 'fascode', 'mipas_1998', and 'mipas_2001'. Default is 'fascode'.
        atm_file : str, optional
            Name of the data file within the chosen folder to load. Default is 'std.atm'.
        interp : str, optional
            One of 'linear' or 'log'. Defines the interpolation space. Default is 'linear'.
        spline: bool, optional
            One of True or False.  If True then a quadratic spline will be used to interpolate, if False the piecewise
            linear interpolation is done.  Default is False.

        Examples
        --------
        >>> from sasktran_tir.climatology import ClimatologyAtmosphericState
        >>> us_standard_atmosphere = ClimatologyAtmosphericState()  # default atmosphere
        >>> std_temperature = us_standard_atmosphere.get_parameter('SKCLIMATOLOGY_TEMPERATURE_K', latitude=0, longitude=0, altitudes=np.linspace(20000, 30000, 4), mjd=54372)
        >>> print(std_temperature)
        [216.7        219.93333333 223.2        226.5       ]
        >>> mipas_2001_polar_summer = ClimatologyAtmosphericState(atm_folder='mipas_2001', atm_file='sum.atm')
        >>> polar_pressure = mipas_2001_polar_summer.get_parameter('SKCLIMATOLOGY_PRESSURE_PA', latitude=0, longitude=0, altitudes=np.linspace(20000, 30000, 4), mjd=54372)
        >>> print(polar_pressure)
        [5537.54       3435.81       2128.99333333 1329.72      ]
        """

        # get climatology data
        file_path = _atm_file_path(atm_folder, atm_file)
        data = _atm_reader(file_path)

        super().__init__(data['HGT'] * 1000.0,
                         {'SKCLIMATOLOGY_TEMPERATURE_K': data['TEM'], 'SKCLIMATOLOGY_PRESSURE_PA': data['PRE'] * 100.0},
                         interp=interp, spline=spline)


class ClimatologySpecies(sk.ClimatologyUserDefined):
    """
    Create a custom climatology for a single species from data files.

    ========== ============= ======================== =================================================================
    Folder     File          Description              Species
    ========== ============= ======================== =================================================================
    fascod     tro.atm       Tropical                 PRE, TEM, H2O, CO2, O3, N2O, CO, CH4, O2
               mls.atm       Mid-Latitude Summer      PRE, TEM, H2O, CO2, O3, N2O, CO, CH4, O2
               mlw.atm       Mid-Latitude Winter      PRE, TEM, H2O, CO2, O3, N2O, CO, CH4, O2
               sas.atm       Sub-Arctic Summer        PRE, TEM, H2O, CO2, O3, N2O, CO, CH4, O2
               saw.atm       Sub-Arctic Winter        PRE, TEM, H2O, CO2, O3, N2O, CO, CH4, O2
               std.atm       US Standard Atmosphere   PRE, TEM, H2O, CO2, O3, N2O, CO, CH4, O2
               minor.atm     additional species       NO, SO2, NO2, NH3, HNO3, OH, HF, HCl, HBr, HI, ClO, OCS,H2CO,
                                                      HOCl, N2, HCN, CH3Cl, H2O2, C2H2, C2H6, PH3, COF2, SF6, H2S,
                                                      CFCl3, CF2Cl2, CClF3, CF4, CHCl2F, CHClF2, C2Cl3F3, C2Cl2F4,
                                                      C2ClF5, CCl4, ClONO2, N2O5, HNO4
               hgt_std.atm   altitude grid only
    ========== ============= ======================== =================================================================
    mipas_1998 day_imk.atm   Mid-Latitude, day-time   PRE, TEM, N2, O2, O3P, CO2, O3, H2O, CH4, N2O, HNO3, CO, NO2,
                                                      N2O5, ClO, HOCl, ClONO2, NO
               ngt_imk.atm   Mid-Latitude, night-time PRE, TEM, N2, O2, O3P, CO2, O3, H2O, CH4, N2O, HNO3, CO, NO2,
                                                      N2O5, ClO, HOCl, ClONO2, NO
               win_imk.atm   Polar Winter             PRE, TEM, N2, O2, O3P, CO2, O3, H2O, CH4, N2O, HNO3, CO, NO2,
                                                      N2O5, ClO, HOCl, ClONO2, NO
               sum_imk.atm   Polar Summer             PRE, TEM, N2, O2, O3P, CO2, O3, H2O, CH4, N2O, HNO3, CO, NO2,
                                                      N2O5, ClO, HOCl, ClONO2, NO
               extra_imk.atm additional species       HCN, H2O2, F12, F14, F22, COF2, OCS, NH3, SO2, CFCl3, C2H2, C2H6,
                                                      CCl4, SF6, HNO4, CH3Cl, CClF3, CHCl2F, C2Cl3F3, C2Cl2F4
    ========== ============= ======================== =================================================================
    mipas_2001 day.atm       Mid-Latitude, day-time   PRE, TEM, N2, O2, CO2, O3, H2O, CH4, N2O, HNO3, CO, NO2, N2O5,
                                                      ClO, HOCl, ClONO2, NO, HNO4, HCN, NH3, F11, F12, F14, F22, CCl4,
                                                      COF2, H2O2, C2H2, C2H6, OCS, SO2, SF6
               ngt.atm       Mid-Latitude, night-time PRE, TEM, N2, O2, CO2, O3, H2O, CH4, N2O, HNO3, CO, NO2, N2O5,
                                                      ClO, HOCl, ClONO2, NO, HNO4, HCN, NH3, F11, F12, F14, F22, CCl4,
                                                      COF2, H2O2, C2H2, C2H6, OCS, SO2, SF6
               win.atm       Polar Winter             PRE, TEM, N2, O2, CO2, O3, H2O, CH4, N2O, HNO3, CO, NO2, N2O5,
                                                      ClO, HOCl, ClONO2, NO, HNO4, HCN, NH3, F11, F12, F14, F22, CCl4,
                                                      COF2, H2O2, C2H2, C2H6, OCS, SO2, SF6
               sum.atm       Polar Summer             PRE, TEM, N2, O2, CO2, O3, H2O, CH4, N2O, HNO3, CO, NO2, N2O5,
                                                      ClO, HOCl, ClONO2, NO, HNO4, HCN, NH3, F11, F12, F14, F22, CCl4,
                                                      COF2, H2O2, C2H2, C2H6, OCS, SO2, SF6
               equ.atm       Equatorial, day-time     PRE, TEM, N2, O2, CO2, O3, H2O, CH4, N2O, HNO3, CO, NO2, N2O5,
                                                      ClO, HOCl, ClONO2, NO, HNO4, HCN, NH3, F11, F12, F14, F22, CCl4,
                                                      COF2, H2O2, C2H2, C2H6, OCS, SO2, SF6
               extra.atm     additional species       CClF3, CHCl2F, C2Cl3F3, C2Cl2F4, C2ClF5, CH3Cl, H2S
    ========== ============= ======================== =================================================================
    """
    def __init__(self, species: str, atm_folder='fascode', atm_file='std.atm', interp: str='linear',
                 spline: bool=False):
        """
        Create a custom climatology using a species profile from data files. By default uses the US Standard Atmosphere.
        The CLIMATOLOGY_HANDLE of the species will be SKCLIMATOLOGY_<species>_CM3. Note that despite the files
        providing profiles of species in PPMV (parts per million by volume) the climatology created will be a profile
        with units of molecules per cm^3. This is because the VMR climatologies are currently not supported in the
        SASKTRAN cross section calculation. The species PPMV is converted to using the temperature, T, and pressure, p,
        in the specified '.atm' file and using the Ideal Gas Law to compute a neutral density, n:

            n = p / (k T)

        where k is the Boltzmann constant.  The species VMR is multiplied by the neutral density to obtain the species
        density.

        NOTE: These are one-dimensional atmospheric profiles. Latitude, longitude, and mjd will not affect the returned
              values.

        Parameters
        ----------
        species : str
            Chemical formula of the species to use. E.g. 'O3', 'CO2', etc.
        atm_folder : str, optional
            Name of dataset to use. Valid choices are 'fascode', 'mipas_1998', and 'mipas_2001'. Default is 'fascode'.
        atm_file : str, optional
            Name of the data file within the chosen folder to load. Default is 'std.atm'.
        interp : str, optional
            One of 'linear' or 'log'. Defines the interpolation space. Default is 'linear'.
        spline: bool, optional
            One of True or False.  If True then a quadratic spline will be used to interpolate, if False the piecewise
            linear interpolation is done.  Default is False.

        Examples
        --------
        >>> import numpy as np
        >>> import sasktran as sk
        >>> from sasktran_tir.climatology import ClimatologySpecies
        >>> mipas_2001_polar_summer_co2 = ClimatologySpecies('CO2', atm_folder='mipas_2001', atm_file='sum.atm')
        >>> n_co2 = mipas_2001_polar_summer_co2.get_parameter('SKCLIMATOLOGY_CO2_CM3', latitude=0, longitude=0, altitudes=np.arange(0, 100000, 500), mjd=54372)
        >>> print(n_co2)
        [6.17180616e+14 3.82011862e+14 2.35914675e+14 1.41142413e+14]
        """

        # get climatology data
        file_path = _atm_file_path(atm_folder, atm_file)
        data = _atm_reader(file_path)

        # Calculate neutral density from Ideal Gas Law, n = p / (k T) then convert from m^-3 to cm^-3
        air_nd = data['PRE'] * 100.0 / Boltzmann / data['TEM'] / 1000000.0

        super().__init__(data['HGT'] * 1000.0,
                         {'SKCLIMATOLOGY_' + species + '_CM3': data[species] / 1.0e6 * air_nd},
                         interp=interp, spline=spline)
